------------------------------------------------------------------------------
--
--  generic package Mandelbrot.Parallel (body)
--
--  A parallel implementation for calculating a map of the Mandelbrot set.
--
------------------------------------------------------------------------------
--  Update information:
--
--  2001.07.30 (Jacob Sparre Andersen)
--    Written.
--
--  2001.08.07 (Jacob Sparre Andersen)
--    Renamed the Job_Server parameter from Line_Count to Height.
--
--  2001.08.08 (Jacob Sparre Andersen)
--    Fixed a mistake in the package description.
--
--  (Insert additional update information above this line.)
------------------------------------------------------------------------------
--  Standard packages:

------------------------------------------------------------------------------
--  Other packages:

------------------------------------------------------------------------------

package body Mandelbrot.Parallel is

   ---------------------------------------------------------------------------
   --  procedure Calculate:

   procedure Calculate (View : in     View_Description;
                        Map  :    out Map_Type) is

      Job         : View_Description_Reference;
      Result      : Result_Buffer_Reference;
      Server      : Job_Server_Reference;
      Calculators : array (1 .. 4) of Calculator_Reference;

   begin --  Calculate
      Job := new View_Description'(View);

      Result := new Result_Buffer  (Width  => View.Width,
                                    Height => View.Height);

      Server := new Job_Server (Height => View.Height);

      for Index in Calculators'Range loop
         Calculators (Index) := new Calculator (Job    => Job,
                                                Server => Server,
                                                Result => Result);
      end loop;

      Result.Complete (Map => Map);
   end Calculate;

   ---------------------------------------------------------------------------
   --  protected type Job_Server:

   protected body Job_Server is
      ------------------------------------------------------------------------
      --  procedure Get:
      --
      --  Returns the number of a line to be processed or "end of job".

      procedure Get (Line       :    out Natural;
                     End_Of_Job :    out Boolean) is

      begin
         if Next_Line > Height then
            End_Of_Job := True;
         else
            Line := Next_Line;
            End_Of_Job := False;

            Next_Line := Next_Line + 1;
         end if;
      end Get;

      ------------------------------------------------------------------------
   end Job_Server;

   ---------------------------------------------------------------------------
   --  protected type Result_Buffer:

   protected body Result_Buffer is
      ------------------------------------------------------------------------
      --  procedure Put:
      --
      --  Writes a line of the map to the result buffer:

      procedure Put (Line : in     Natural;
                     Data : in     Line_Type) is

      begin --  Put
         if Filled_Lines (Line) then
            null; --  Could verify correctness here.
         else
            for Column in Data'Range loop
               Pixmap (Column, Line) := Data (Column);
               --  Assumes that Data'Range = Pixmap'Range (1)!!!
            end loop;

            Filled_Lines (Line) := True;
            Remaining_Lines := Remaining_Lines - 1;
         end if;
      end Put;

      ------------------------------------------------------------------------
      --  entry Complete:
      --
      --  Available when all lines have been written to the result buffer.

      entry Complete (Map :    out Map_Type) when Remaining_Lines = 0 is

      begin --  Complete
         Map := Pixmap;
      end Complete;

      ------------------------------------------------------------------------
   end Result_Buffer;

   ---------------------------------------------------------------------------
   --  task type Calculator:

   task body Calculator is

      use Complex_Types;

      use type Mandelbrot.Pixel_Type;

      Done             : Boolean;
      X_Index, Y_Index : Positive;
      Re_Start         : Real;
      Im_Start         : Imaginary;
      Z_0, Z           : Complex;
      Data             : Line_Type (1 .. Job.Width);

   begin --  Calculator
      loop
         Server.Get (Line       => Y_Index,
                     End_Of_Job => Done);

         exit when Done;

         Im_Start := Job.Im_Min + (Job.Im_Max - Job.Im_Min) *
                       Real (Y_Index - 1) / Real (Job.Height);

      Scan_Line:
         for Column in Data'Range  loop
            Re_Start := Job.Re_Min + (Job.Re_Max - Job.Re_Min) *
                          Real (Column - 1) / Real (Job.Width);

            Z_0 := Re_Start + Im_Start;
            Z   := Z_0;

            Data (Column) := Pixel_Type (Job.Iteration_Limit) + 1;

         Iterate:
            for Iteration in 0 .. Job.Iteration_Limit loop
               if Modulus (Z) > 2.0 then
                  Data (Column) := Pixel_Type (Iteration);
                  exit Iterate;
               end if;

               Z := Z_0 + Z ** 2; --  Tjek den lige!
            end loop Iterate;
         end loop Scan_Line;

         Result.Put (Line => Y_Index,
                     Data => Data);
      end loop;
   end Calculator;

   ---------------------------------------------------------------------------

end Mandelbrot.Parallel;
